@@ -2,13 +2,16 @@ import sys, os
 from PyQt6.QtCore import Qt, QModelIndex, QSettings, QUrl, QEvent
 from PyQt6.QtGui import QFontMetrics, QDesktopServices, QWindow
 from PyQt6.QtWidgets import *
-from .die import DIETableModel, ip_in_range
-from .formats import read_dwarf
-from .tree import DWARFTreeModel, has_code_location, cu_sort_key
-from .scriptdlg import ScriptDlg
+from .die import DIETableModel
+from .formats import read_dwarf, get_debug_sections, FormatError
+from .dwarfutil import has_code_location, ip_in_range
+from .tree import DWARFTreeModel, cu_sort_key
+from .scriptdlg import ScriptDlg, make_execution_environment
 from .ui import setup_ui
+from .locals import LocalsDlg
 
-version = (2, 31)
+# Sync with version in setup.py
+version = (4, 0)
 
 # TODO:
 # On MacOS, start without a main window, instead show the Open dialog
@@ -18,6 +21,12 @@ version = (2, 31)
 # Pretty much DWARF unaware, all the DWARF visualization logic is in tree.py and die.py
 #-----------------------------------------------------------------
 
+# "Opened, could not parse"
+class DWARFParseError(Exception):
+    def __init__(self, exc, di):
+        Exception.__init__(self, "DWARF parsing error: " + format(exc))
+        self.dwarfinfo = di
+
 class TheWindow(QMainWindow):
     def __init__(self):
         QMainWindow.__init__(self)
@@ -68,10 +77,11 @@ class TheWindow(QMainWindow):
 
     # Callback for the Mach-O fat binary opening logic
     # Taking a cue from Hopper or IDA, we parse only one slice at a time
-    def resolve_arch(self, arches):
-        r = QInputDialog.getItem(self, 'Mach-O Fat Binary', 'Choose an architecture:', arches, 0, False, Qt.WindowType.Dialog)
+    # Also reused for 
+    def resolve_arch(self, arches, title, message):
+        r = QInputDialog.getItem(self, title, message, arches, 0, False, Qt.WindowType.Dialog)
         return arches.index(r[0]) if r[1] else None
-
+    
     # Can throw an exception
     # Returns None if it doesn't seem to contain DWARF
     # False if the user cancelled
@@ -79,13 +89,24 @@ class TheWindow(QMainWindow):
     def open_file(self, filename, arch = None):
         self.start_wait()
         try:
-            di = read_dwarf(filename, self.resolve_arch if arch is None else lambda arches: arches.index(arch))
+            di = read_dwarf(filename, self.resolve_arch if arch is None else lambda arches, title, text: arches.index(arch))
             if not di: # Covers both False and None
                 return di
+            
+            return self.load_dwarfinfo(di, filename, arch)
+        finally:
+            self.end_wait()
 
+    # May throw if parsing fails
+    def load_dwarfinfo(self, di, filename, arch):
+        # Arch comes from opening with arch (via MRU), not from the interactive selection
+        # Some degree of graceful handling of wrong format
+        try:
+            slice = di._fat_arch if hasattr(di, '_fat_arch') else None
             # Some cached top level stuff
-            # Notably, iter_CUs doesn't cache
+            # Notably, iter_CUs doesn't cache (TODO, check that in the next version)
             di._ranges = None # Loaded on first use
+            di._aranges = None
             def decorate_cu(cu, i):
                 cu._i = i
                 cu._lineprogram = None
@@ -102,36 +123,51 @@ class TheWindow(QMainWindow):
                 di._CUs.sort(key = cu_sort_key)
                 for (i, cu) in enumerate(di._CUs):
                     cu._i = i
-            di._locparser = None # Created on first use
+            di._locparser = None # Created on first use - but see #1683
 
             self.dwarfinfo = di
+            self.filename = filename
             self.tree_model = DWARFTreeModel(di, self.prefix, self.sortcus, self.sortdies)
             self.the_tree.setModel(self.tree_model)
             self.the_tree.selectionModel().currentChanged.connect(self.on_tree_selection)
             s = os.path.basename(filename)
-            if arch is not None:
-                s += ' (' + arch + ')'
+            if slice is not None:
+                s += ' (' + slice + ')'
             self.setWindowTitle("DWARF Explorer - " + s)
+            self.savesection_menuitem.setEnabled(True)
+            self.switchslice_menuitem.setEnabled(slice is not None)
             self.back_menuitem.setEnabled(False)
+            self.back_tbitem.setEnabled(False)
             self.forward_menuitem.setEnabled(False)
+            self.forward_tbitem.setEnabled(False)
             self.followref_menuitem.setEnabled(False)
+            self.followref_tbitem.setEnabled(False)
             self.highlightcode_menuitem.setEnabled(True)
+            self.highlightsubstring_menuitem.setEnabled(True)
+            self.highlightcondition_menuitem.setEnabled(True)
             self.highlightnothing_menuitem.setEnabled(True)
             self.copy_menuitem.setEnabled(False)
+            self.copy_tbitem.setEnabled(False)
             self.copyline_menuitem.setEnabled(False)
             self.copytable_menuitem.setEnabled(False)
             self.findbycondition_menuitem.setEnabled(True)
             self.find_menuitem.setEnabled(True)
+            self.find_tbitem.setEnabled(True)
             self.findip_menuitem.setEnabled(True)
             self.byoffset_menuitem.setEnabled(True)
+            self.byoffset_tbitem.setEnabled(True)
+            self.localsat_menuitem.setEnabled(True)
             self.on_highlight_nothing()
             # Navigation stack - empty
             self.navhistory = []
             self.navpos = -1
-            self.save_filename_in_mru(filename, di._fat_arch if '_fat_arch' in dir(di) and di._fat_arch else None)
+            self.save_filename_in_mru(filename, slice)
+            LocalsDlg.reset(di)
+            from .crash import set_binary_desc
+            set_binary_desc(("ELF", "MachO", "PE", "WASM", "ELFinA", "MachOinA")[di._format] + " " + di.config.machine_arch)
             return True
-        finally:
-            self.end_wait()
+        except AssertionError as ass: # Covers exeptions during parsing
+            raise DWARFParseError(ass, di)        
 
     def save_mru(self):
         for i, fa in enumerate(self.mru):
@@ -140,6 +176,9 @@ class TheWindow(QMainWindow):
                 self.sett.setValue("General/MRUArch%d" % i, fa[1])
             else:
                 self.sett.remove("General/MRUArch%d" % i)
+        for i in range(len(self.mru), 10):
+            self.sett.remove("General/MRU%d" % i)
+            self.sett.remove("General/MRUArch%d" % i)
 
     # Open a file, display an error if failure
     def open_file_interactive(self, filename, arch = None):
@@ -151,10 +190,49 @@ class TheWindow(QMainWindow):
                     s = "The file contains no DWARF information, or it is in an unsupported format."
                 QMessageBox(QMessageBox.Icon.Warning, "DWARF Explorer", s,
                     QMessageBox.StandardButton.Ok, self).show()
+        except FormatError as ferr:
+            QMessageBox(QMessageBox.Icon.Warning, "DWARF Explorer",
+                str(ferr), QMessageBox.StandardButton.Ok, self).show()
+        except DWARFParseError as dperr:
+            mb = QMessageBox(QMessageBox.Icon.Critical, "DWARF Explorer",
+                "Error parsing the DWARF information in this file. Would you like to save the debug section contents for manual analysis?",
+                QMessageBox.StandardButton.Yes|QMessageBox.StandardButton.No, self)
+            mb.setEscapeButton(QMessageBox.StandardButton.No)
+            r = mb.exec()
+            if r == QMessageBox.StandardButton.Yes:
+                self.save_sections(filename, dperr.dwarfinfo)
         except Exception as exc:
             QMessageBox(QMessageBox.Icon.Critical, "DWARF Explorer",
                 "Error opening the file:\n\n" + format(exc),
                 QMessageBox.StandardButton.Ok, self).show()
+            
+    def save_sections(self, filename, di):
+        dir = QFileDialog.getExistingDirectory(self, "Choose a save location", os.path.dirname(filename))
+        if dir:
+            sections = get_debug_sections(di)
+            basename = os.path.basename(filename)
+            overwrite_all = False
+            for (name, section) in sections.items():
+                try:
+                    section_file = os.path.join(dir, basename + '.' + name)
+                    skip = False
+                    if os.path.exists(section_file) and not overwrite_all:
+                        mb = QMessageBox(QMessageBox.Icon.Question, "DWARF Explorer",
+                            "File %s exists, overwrite?" % section_file,
+                            QMessageBox.StandardButton.Yes|QMessageBox.StandardButton.YesAll|QMessageBox.StandardButton.No|QMessageBox.StandardButton.Cancel, self)
+                        mb.setEscapeButton(QMessageBox.StandardButton.Cancel)
+                        r = mb.exec()
+                        if r == QMessageBox.StandardButton.Cancel:
+                            return
+                        elif r == QMessageBox.StandardButton.YesAll:
+                            overwrite_all = True
+                        elif r == QMessageBox.StandardButton.No:
+                            skip = True
+                    if not skip:
+                        with open(section_file, 'wb') as f:
+                            f.write(section.stream.getbuffer())
+                except:
+                    pass
 
     # TODO: list the extensions for the open file dialog?
     def on_open(self):
@@ -170,7 +248,17 @@ class TheWindow(QMainWindow):
                 self.fa = fa
                 self.win = win
             def __call__(self):
-                self.win.open_file_interactive(*self.fa)
+                fa = self.fa
+                if os.path.exists(self.fa[0]):
+                    self.win.open_file_interactive(*fa)
+                else:
+                    mb =  QMessageBox(QMessageBox.Icon.Critical, "DWARF Explorer",
+                        "The file or bundle %s does not exist or is not accessible. Shall we remove it from the recent file menu?" % (fa[0],),
+                        QMessageBox.StandardButton.Yes|QMessageBox.StandardButton.No, self.win)
+                    mb.setEscapeButton(QMessageBox.StandardButton.No)
+                    r = mb.exec()
+                    if r == QMessageBox.StandardButton.Yes:
+                        self.win.delete_from_mru(fa)
 
         for i, fa in enumerate(self.mru):
             s = fa[0]
@@ -193,7 +281,17 @@ class TheWindow(QMainWindow):
             self.save_mru()
             self.mru_menu.setEnabled(True)
             self.mru_menu.clear()
-            self.populate_mru_menu()    
+            self.populate_mru_menu()
+
+    def delete_from_mru(self, mru_record):
+        try:
+            self.mru.remove(mru_record) # ValueError if not found
+            self.save_mru()
+            self.mru_menu.setEnabled(len(self.mru) > 0)
+            self.mru_menu.clear()
+            self.populate_mru_menu()
+        except ValueError:
+            pass
 
     # File drag/drop handling - equivalent to open
     def dragEnterEvent(self, evt):
@@ -201,7 +299,34 @@ class TheWindow(QMainWindow):
             evt.accept()
 
     def dropEvent(self, evt):
-        self.open_file_interactive(os.path.normpath(evt.mimeData().urls()[0].toLocalFile()))               
+        self.open_file_interactive(os.path.normpath(evt.mimeData().urls()[0].toLocalFile()))
+
+    # Save sections as
+    def on_savesection(self):
+        di = self.dwarfinfo
+        # Maps display name to field name in DWARFInfo
+        sections = get_debug_sections(di)
+        
+        names = sections.keys()
+        r = QInputDialog.getItem(self, 'Save a Section', 'Choose a section:', names, 0, False, Qt.WindowType.Dialog)
+        if r[1]:
+            section_name = r[0]
+            r = QFileDialog.getSaveFileName(self, "Save a section", self.filename + '.' + section_name)
+            if r[0]:
+                try:
+                    section = sections[section_name]
+                    with open(r[0], 'wb') as f:
+                        # Assumes the section is a BytesIO - implementation dependent
+                        f.write(section.stream.getbuffer())
+                except Exception as exc:
+                    QMessageBox(QMessageBox.Icon.Critical, "DWARF Explorer",
+                        "Error saving the section data:\n\n" + format(exc),
+                        QMessageBox.StandardButton.Ok, self).show()
+    
+    # Just present the slice dialog again
+    # TODO: do this in a more elegant way, without reopening and rereading
+    def on_switchslice(self):
+        self.open_file(self.filename, None)
 
     #############################################################
     # Done with file stuff, now tree navigation
@@ -220,6 +345,7 @@ class TheWindow(QMainWindow):
         self.die_table.resizeColumnsToContents()
         self.details_table.setModel(None)
         self.followref_menuitem.setEnabled(False)
+        self.followref_tbitem.setEnabled(False)
         self.cuproperties_menuitem.setEnabled(True)
         self.die_table.setCurrentIndex(QModelIndex()) # Will cause on_attribute_selection
 
@@ -239,7 +365,9 @@ class TheWindow(QMainWindow):
                 self.navhistory[0:self.navpos] = [navitem]
                 self.navpos = 0
                 self.back_menuitem.setEnabled(len(self.navhistory) > 1)
+                self.back_tbitem.setEnabled(len(self.navhistory) > 1)
                 self.forward_menuitem.setEnabled(False)
+                self.forward_tbitem.setEnabled(False)
         self.display_die(index) # Will clear the selection in the attribute table
 
     # Selection changed in the DIE table - either user or program
@@ -255,15 +383,19 @@ class TheWindow(QMainWindow):
             else:
                 self.details_warning.setVisible(False)
             self.followref_menuitem.setEnabled(self.die_model.ref_target(index) is not None)
+            self.followref_tbitem.setEnabled(self.die_model.ref_target(index) is not None)
             self.copy_menuitem.setEnabled(True)
+            self.copy_tbitem.setEnabled(True)
             self.copyline_menuitem.setEnabled(True)
             self.copytable_menuitem.setEnabled(True)
         else: # Selected nothing
             self.details_table.setModel(None)
             self.copy_menuitem.setEnabled(False)
+            self.copy_tbitem.setEnabled(False)
             self.copyline_menuitem.setEnabled(False)
             self.copytable_menuitem.setEnabled(False)            
             self.followref_menuitem.setEnabled(False)
+            self.followref_tbitem.setEnabled(False)
 
 
     def on_attribute_dclick(self, index):
@@ -283,7 +415,9 @@ class TheWindow(QMainWindow):
             self.the_tree.setCurrentIndex(tree_index) # Causes on_tree_selection internally
             self.in_tree_nav = False
             self.back_menuitem.setEnabled(np < len(self.navhistory) - 1)
+            self.back_tbitem.setEnabled(np < len(self.navhistory) - 1)
             self.forward_menuitem.setEnabled(np > 0)
+            self.forward_tbitem.setEnabled(np > 0)
 
     def followref(self, index = None):
         self.start_wait() # TODO: only show the wait cursor if it's indeed time consuming
@@ -328,13 +462,15 @@ class TheWindow(QMainWindow):
     # Exception means false
     def eval_user_condition(self, cond, die):
         try:
-            def has_attribute(func):
-                for k in die.attributes:
-                    if func(k, die.attributes[k].value, die.attributes[k].form):
-                        return True
-
-            return eval(cond, {'die' : die, 'has_attribute' : has_attribute})
-        except Exception as exc:
+            env = make_execution_environment(die)
+        except Exception as exc: # Our error
+            from .crash import report_crash
+            from inspect import currentframe
+            report_crash(exc, exc.__traceback__, version, currentframe())
+            return False
+        try:
+            return eval(cond, env)
+        except Exception as exc: # Error in condition or it assumes a different DIE structure 
             print("Error in user condition: %s" % format(exc))
             return False
 
@@ -367,10 +503,7 @@ class TheWindow(QMainWindow):
         r = QInputDialog.getText(self, "Find DIE by offset", "DIE offset (hex), relative to the section start:")
         if r[1] and r[0]:
             try:
-                offset = r[0]
-                if offset.startswith("0x"):
-                    offset = offset[2:]
-                offset = int(offset, 16)
+                offset = int(r[0], 16)
                 index = self.tree_model.find_offset(offset)
                 if index:
                     self.the_tree.setCurrentIndex(index)
@@ -380,9 +513,11 @@ class TheWindow(QMainWindow):
             except ValueError:
                 pass
 
+    def sample_die(self):
+        return self.the_tree.currentIndex().internalPointer() or self.dwarfinfo._CUs[0].get_top_DIE()
 
     def on_findbycondition(self):
-        dlg = ScriptDlg(self)
+        dlg = ScriptDlg(self, self.sample_die())
         if dlg.exec() == QDialog.DialogCode.Accepted:
             cond = dlg.cond
             self.findcondition = lambda die: self.eval_user_condition(cond, die)
@@ -399,7 +534,7 @@ class TheWindow(QMainWindow):
     ##########################################################################
 
     def on_about(self):
-        QMessageBox(QMessageBox.Icon.Information, "About...", "DWARF Explorer v." + '.'.join(str(v) for v in version) + "\n\nSeva Alekseyev, 2020-2022\nsevaa@sprynet.com\n\ngithub.com/sevaa/dwex",
+        QMessageBox(QMessageBox.Icon.Information, "About...", "DWARF Explorer v." + '.'.join(str(v) for v in version) + "\n\nSeva Alekseyev, 2020-2023\nsevaa@sprynet.com\n\ngithub.com/sevaa/dwex",
             QMessageBox.StandardButton.Ok, self).show()
 
     def on_updatecheck(self):
@@ -478,21 +613,79 @@ class TheWindow(QMainWindow):
             sel = self.tree_model.set_sortdies(checked)
             #This invalidates the navigation
             self.back_menuitem.setEnabled(False)
+            self.back_tbitem.setEnabled(False)
             self.forward_menuitem.setEnabled(False)
+            self.forward_tbitem.setEnabled(False)
             self.followref_menuitem.setEnabled(False)
+            self.followref_tbitem.setEnabled(False)
             # Navigation stack - empty
             self.navhistory = []
             self.navpos = -1
             if sel:
                 self.the_tree.setCurrentIndex(sel)
 
+    # Tree highlighting business
+
+    def manage_hlnavigation(self, b = None):
+        if b is None:
+            b = self.tree_model.has_any_highlights()
+        self.prevhl_menuitem.setEnabled(b)
+        self.nexthl_menuitem.setEnabled(b)
+        self.prevhl_tbitem.setEnabled(b)
+        self.nexthl_tbitem.setEnabled(b)
+
+    def highlight_off(self, key):
+        self.tree_model.remove_highlight(key)
+        self.manage_hlnavigation()
+
     def on_highlight_code(self):
-        self.highlightcode_menuitem.setChecked(True)
-        self.tree_model.highlight(has_code_location)
+        if self.tree_model.has_highlight(1):
+            self.highlight_off(1)
+        else:
+            self.tree_model.add_highlight(1, has_code_location)        
+            self.manage_hlnavigation(True)
+
+    def on_highlight_substring(self):
+        if self.tree_model.has_highlight(2):
+            self.highlight_off(2)
+        else:
+            r = QInputDialog.getText(self, 'Highlight', 'Highlight DIEs with substring:')
+            if r[1] and r[0]:
+                s = r[0].lower()
+                self.tree_model.add_highlight(2, lambda die:self.findbytext(die, s))
+                self.manage_hlnavigation(True)
+            else:
+                self.highlightsubstring_menuitem.setChecked(False)
+
+    def on_highlight_condition(self):
+        if self.tree_model.has_highlight(3):
+            self.highlight_off(3)
+        else:
+            dlg = ScriptDlg(self, self.sample_die())
+            if dlg.exec() == QDialog.DialogCode.Accepted:
+                cond = dlg.cond
+                self.tree_model.add_highlight(3, lambda die: self.eval_user_condition(cond, die))
+                self.manage_hlnavigation(True)
+            else:
+                self.highlightcondition_menuitem.setChecked(False)
+            # Accepted with blank or bogus expression is not supported
 
     def on_highlight_nothing(self):
         self.highlightcode_menuitem.setChecked(False)
-        self.tree_model.highlight(None)
+        self.highlightsubstring_menuitem.setChecked(False)
+        self.highlightcondition_menuitem.setChecked(False)
+        self.manage_hlnavigation(False)
+        self.tree_model.clear_highlight()
+
+    def on_nexthl(self):
+        index = self.tree_model.find(self.the_tree.currentIndex(), self.tree_model.is_highlighted, False)
+        if index:
+            self.the_tree.setCurrentIndex(index)
+
+    def on_prevhl(self):
+        index = self.tree_model.find_back(self.the_tree.currentIndex(), self.tree_model.is_highlighted, False)
+        if index:
+            self.the_tree.setCurrentIndex(index)
 
     def on_cuproperties(self):
         die = self.the_tree.currentIndex().internalPointer()
@@ -535,6 +728,11 @@ class TheWindow(QMainWindow):
             for r in range(0, m.rowCount(QModelIndex())))
         self.on_copy(table_text)
 
+    def on_localsat(self):
+        dlg = LocalsDlg(self, self.dwarfinfo, self.prefix, self.dwarfregnames, self.hex)
+        if dlg.exec() == QDialog.DialogCode.Accepted and dlg.selected_die:
+             self.the_tree.setCurrentIndex(self.tree_model.index_for_die(dlg.selected_die))
+
     # If the details pane has data - reload that
     def refresh_details(self):
         index = self.die_table.currentIndex()
@@ -551,7 +749,51 @@ class TheWindow(QMainWindow):
     def on_homepage(self):
         QDesktopServices.openUrl(QUrl('https://github.com/sevaa/dwex'))
 
-    # Doesn't quite work for the delay on tree expansion :(
+    # All purpose debug hook
+    def on_debug(self):
+        self.dwarfinfo.CFI_entries()
+        return
+
+
+        import io
+        from elftools.dwarf.dwarfinfo import DWARFInfo, DwarfConfig, DebugSectionDescriptor
+        # Read the three saved sections as bytestreams
+        base = os.environ.get("DWEX_ARG")
+        with open(base + '.info.dat', 'rb') as f:
+            info = f.read()
+        with open(base + '.abbrev.dat', 'rb') as f:
+            abbrev = f.read()
+        with open(base + '.str.dat', 'rb') as f:
+            str = f.read()
+
+        # Parse the DWARF info
+        di = DWARFInfo(
+            config = DwarfConfig(little_endian = True, default_address_size = 8, machine_arch = "ARM64"),
+            debug_info_sec = DebugSectionDescriptor(io.BytesIO(info), '__debug_info', None, len(info), 0),
+            debug_aranges_sec = None,
+            debug_abbrev_sec = DebugSectionDescriptor(io.BytesIO(abbrev), '__debug_abbrev', None, len(abbrev), 0),
+            debug_frame_sec = None,
+            eh_frame_sec = None,
+            debug_str_sec = DebugSectionDescriptor(io.BytesIO(str), '__debug_str', None, len(str), 0),
+            debug_loc_sec = None,
+            debug_ranges_sec = None,
+            debug_line_sec = DebugSectionDescriptor(io.BytesIO(b''), '__debug_line', None, 0, 0),
+            debug_pubtypes_sec = None,
+            debug_pubnames_sec = None,
+            debug_addr_sec=None,
+            debug_str_offsets_sec=None,
+            debug_line_str_sec=None,
+            debug_loclists_sec = None,
+            debug_rnglists_sec = None,
+            debug_sup_sec = None,
+            gnu_debugaltlink_sec = None
+        )
+        di._start_address = 0
+        di._format = 1
+
+        self.load_dwarfinfo(di, "", "ARM64")
+
+    # Doesn't quite work for the delay on tree expansion :( TODO: timer checks before lighting up this
     def start_wait(self):
         QApplication.setOverrideCursor(Qt.CursorShape.WaitCursor)
 
@@ -559,9 +801,16 @@ class TheWindow(QMainWindow):
         QApplication.restoreOverrideCursor()
 
 def on_exception(exctype, exc, tb):
-    if isinstance(exc, Exception):
+    if isinstance(exc, MemoryError):
+        app = QApplication.instance()
+        app.win.destroy()
+        app.win = None
+        QMessageBox(QMessageBox.Icon.Critical, "DWARF Explorer", "Out of memory. DWARF Explorer will now close. Sorry.",
+            QMessageBox.StandardButton.Ok, None).show()
+        sys.exit(1)
+    elif isinstance(exc, Exception):
         from .crash import report_crash
-        report_crash(exc, tb, version)
+        report_crash(exc, exc.__traceback__, version)
         sys.excepthook = on_exception.prev_exchook
         sys.exit(1)
     elif on_exception.prev_exchook:
@@ -584,8 +833,9 @@ class TheApp(QApplication):
         self.win = TheWindow()
         self.exec()
 
-def main():     
-    if not sys.gettrace(): # Lame way to detect a debugger
+def main():
+    under_debugger = sys.gettrace() # Lame way to detect a debugger
+    if not under_debugger: 
         on_exception.prev_exchook = sys.excepthook
         sys.excepthook = on_exception
 
