@@ -1,5 +1,12 @@
 import hashlib, urllib, random, string, traceback, sys, os, platform
 from urllib.request import urlopen
+# No dependencies on the rest of the app, and keep it that way
+
+_binary_desc = None
+
+def set_binary_desc(s):
+    global _binary_desc
+    _binary_desc = s
 
 def submit_report(subj, body):
     try:
@@ -44,27 +51,55 @@ def get_crash_die_context(locals, ctxt = None):
         pass
     return s
 
-def make_exc_report(exc, tb, version, ctxt=None):
+def make_exc_report(exc, tb, version, catchpoint, ctxt=None):
+    global _binary_desc
+    tracebacks = []
     while tb.tb_next:
+        tracebacks.insert(0, tb) # Innermost in the beginning of the list
         tb = tb.tb_next
     ss = traceback.extract_tb(tb)
-    stack = traceback.extract_stack(tb.tb_frame)
+    innermost_stack = traceback.extract_stack(tb.tb_frame)
     crashpoint = ss[0]
     locals = tb.tb_frame.f_locals
 
     report = type(exc).__name__ + ' at ' + os.path.basename(crashpoint.filename) + ':' + str(crashpoint.lineno) + "\n"
+    if catchpoint:
+        from inspect import getframeinfo
+        fi = getframeinfo(catchpoint)
+        report += "Caught %s@%s:%d\n" % (fi.function, os.path.basename(fi.filename), fi.lineno)
     report += "DWEX " + '.'.join(str(v) for v in version) + "\n"
     report += "Python " + sys.version + "\n"
-    report +=  platform.platform() + "\n"
-    from .cookie import cookie
+    report += "System: " + platform.platform() + "\n"
+    if _binary_desc:
+        report += "Binary: " + _binary_desc + "\n"
+    try:
+        from .cookie import cookie
+    except ImportError:
+        cookie = False
     if cookie:
         report += "Cookie: " + cookie + "\n"
     report += get_crash_die_context(locals, ctxt=ctxt)
     report += "".join(traceback.format_exception_only(type(exc), exc)) + "\n"
 
-    report += "PyStack:\n"
-    stacklines = [se.name + ' (' + os.path.basename(se.filename) + ':' + str(se.lineno) + ")\n" for se in stack]
-    report += "".join(stacklines[::-1]) + "\n"
+    report += "PyStack_v3:\n"
+    def module_prefix(se):
+        p = os.path.dirname(se.filename).split(os.path.sep)
+        if 'elftools' in p:
+            return 'pyelftools/'
+        elif 'dwex' in p:
+            return 'dwex/'
+        return ''
+    def make_stackline(se):
+        return se.name + '@' + module_prefix(se) + os.path.basename(se.filename) + ':' + str(se.lineno) + "\n"
+    def make_stack_dump(stack):
+        return [make_stackline(se) for se in stack[::-1]]
+    def make_traceback_dump(tb):
+        return "-\n"+"".join(make_stack_dump(traceback.extract_stack(tb.tb_frame)))
+
+    report += "".join(make_stack_dump(innermost_stack))
+    # More tracebacks
+    report += "".join(make_traceback_dump(tb) for tb in tracebacks[1:])
+    report += "\n"
 
     report += "PyLocals:\n" + ''.join(k + ": " + str(locals[k]) + "\n" for k in locals).replace("\n\n","\n")
 
@@ -73,9 +108,12 @@ def make_exc_report(exc, tb, version, ctxt=None):
 
     return report
 
-def report_crash(exc, tb, version, ctxt=None):
+def report_crash(exc, tb, version, catchpoint = None, ctxt=None):
     try:
-        submit_report('[crash][python][dwex][pyexception]', make_exc_report(exc, tb, version, ctxt=ctxt))
+        subj = '[python][dwex][pyexception]'
+        if not catchpoint:
+            subj += '[crash]'
+        submit_report(subj, make_exc_report(exc, tb, version, catchpoint, ctxt=ctxt))
     except Exception:
         pass
 
@@ -91,4 +129,5 @@ if __name__ == "__main__":
 
         foo()
     except Exception as exc:
-        submit_report(exc, (0, 50))
+        from inspect import currentframe
+        report_crash(exc, exc.__traceback__, (0, 50), currentframe())
