@@ -1,60 +1,21 @@
 from bisect import bisect_right
-from PyQt6.QtCore import Qt, QAbstractItemModel, QAbstractTableModel, QModelIndex
-from PyQt6.QtGui import QBrush, QFont
+from PyQt6.QtCore import Qt, QAbstractTableModel, QModelIndex
+from PyQt6.QtGui import QBrush
 from elftools.dwarf.locationlists import LocationParser, LocationExpr
-from elftools.dwarf.dwarf_expr import DWARFExprParser, DWARFExprOp, DW_OP_opcode2name
-from elftools.dwarf.descriptions import _DESCR_DW_LANG, _DESCR_DW_ATE, _DESCR_DW_ACCESS, _DESCR_DW_INL, _REG_NAMES_x86, _REG_NAMES_x64
+from elftools.dwarf.dwarf_expr import DWARFExprParser
+from elftools.dwarf.descriptions import _DESCR_DW_LANG, _DESCR_DW_ATE, _DESCR_DW_ACCESS, _DESCR_DW_INL, _DESCR_DW_CC
 from elftools.common.exceptions import ELFParseError
-from elftools.dwarf.ranges import BaseAddressEntry as RangeBaseAddressEntry, RangeEntry
+
+from dwex.exprutil import ExprFormatter
 from .dwarfone import DWARFExprParserV1
+from .dwarfutil import *
+from .details import GenericTableModel, FixedWidthTableModel
 
 MAX_INLINE_BYTEARRAY_LEN = 32
 
-_REG_NAMES_ARM = [
-    'r0', 'r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7',
-    'r8', 'r9', 'r10', 'r11', 'r12', 'sp', 'lr', 'pc'
-    # Ever higher values for FP or SIMD registers?
-]
-
-# TODO: check against readelf, not sure
-_REG_NAMES_ARM64 = [
-    'x0', 'x1', 'x2', 'x3', 'x4', 'x5', 'x6', 'x7',
-    'x8', 'x9', 'x10', 'x11', 'x12', 'x13', 'x14', 'x15',
-    'x16', 'x17', 'x18', 'x19', 'x20', 'x21', 'x22', 'x23',
-    'x24', 'x25', 'x26', 'x27', 'x28', 'x29', 'x30', 'x31'
-]
-
-_REG_NAMES_MIPS = [
-    '$zero', '$at', '$v0', '$v1', '$a0', '$a1', '$a2', '$a3',
-    '$t0', '$t1', '$t2', '$t3', '$t4', '$t5', '$t6', '$t7',
-    '$s0', '$s1', '$s2', '$s3', '$s4', '$s5', '$s6', '$s7',
-    '$t8', '$t9', '$k0', '$k1', '$gp', '$sp', '$fp', '$ra'
-] # Lo, hi???
-
-# More? 
-
-# The key here is the machine_arch value in the DWARFConfig struct.
-# Machine arch values are generated differently for ELF, MachO and PE
-# For ELF, see the values in the architecture dict in get_machine_arch() under elftools.elf.elffile
-# For PE, see IMAGE_FILE_MACHINE in filebytes.pe
-# For MachO, see make_macho_arch_name() in formats.py
-
-_REG_NAME_MAP = dict(
-    x86 = _REG_NAMES_x86,
-    I386 = _REG_NAMES_x86,
-    x64 = _REG_NAMES_x64,
-    AMD64 = _REG_NAMES_x64,
-    X86_64 = _REG_NAMES_x64,
-    ARM = _REG_NAMES_ARM,
-    ARMV6 = _REG_NAMES_ARM,
-    ARMV7 = _REG_NAMES_ARM,
-    ARMV7A = _REG_NAMES_ARM,
-    ARMV7S = _REG_NAMES_ARM,
-    AArch64 = _REG_NAMES_ARM64,
-    ARM64 = _REG_NAMES_ARM64,
-    ARME = _REG_NAMES_ARM64,
-    MIPS = _REG_NAMES_MIPS
-)
+def is_long_blob(attr):
+    val = attr.value
+    return ((isinstance(val, bytes) and attr.form not in ('DW_FORM_strp', 'DW_FORM_string')) or is_int_list(val)) and len(val) > MAX_INLINE_BYTEARRAY_LEN
 
 #------------------------------------------------
 # DIE formatter
@@ -62,57 +23,12 @@ _REG_NAME_MAP = dict(
 
 _blue_brush = QBrush(Qt.GlobalColor.blue)
 _ltgrey_brush = QBrush(Qt.GlobalColor.lightGray)
-_fixed_font = None
 
 _ll_headers = ("Attribute", "Offset", "Form", "Raw", "Value")
 _noll_headers = ("Attribute", "Form", "Value")
 _meta_desc = ('DIE offset', 'DIE size', 'Abbrev code', 'Has children') # Anything else?
 _meta_count = 4 # Extra rows if low level detail showing is set
 
-class NoBaseError(Exception):
-    pass
-
-#May return None!
-def get_cu_base(die):
-    top_die = die.cu.get_top_DIE()
-    if 'DW_AT_low_pc' in top_die.attributes:
-        return top_die.attributes['DW_AT_low_pc'].value
-    elif 'DW_AT_entry_pc' in top_die.attributes:
-        return top_die.attributes['DW_AT_entry_pc'].value
-    # TODO: ranges?
-    elif 'DW_AT_ranges' in top_die.attributes:
-        di = die.dwarfinfo
-        if not di._ranges:
-            di._ranges = di.range_lists()
-        if not di._ranges: # Absent in the DWARF file
-            raise NoBaseError()
-        rl = di._ranges.get_range_list_at_offset(top_die.attributes['DW_AT_ranges'].value, cu=die.cu)
-        base = None
-        for r in rl:
-            if isinstance(r, RangeBaseAddressEntry) and (base is None or r.base_address < base):
-                base = r.base_address
-        if base is None:
-            raise NoBaseError()
-        return base
-    else:
-        raise NoBaseError()
-
-def is_int_list(val):
-    return isinstance(val, list) and len(val) > 0 and isinstance(val[0], int)
-
-def is_long_blob(attr):
-    val = attr.value
-    return ((isinstance(val, bytes) and attr.form not in ('DW_FORM_strp', 'DW_FORM_string')) or is_int_list(val)) and len(val) > MAX_INLINE_BYTEARRAY_LEN
-
-def is_block(form):
-    return form in ('DW_FORM_block', 'DW_FORM_block1', 'DW_FORM_block2', 'DW_FORM_block4')
-
-def DIE_name(die):
-    return die.attributes['DW_AT_name'].value.decode('utf-8', errors='ignore')
-
-def safe_DIE_name(die, default = ''):
-    return die.attributes['DW_AT_name'].value.decode('utf-8', errors='ignore') if 'DW_AT_name' in die.attributes else default
-
 class DIETableModel(QAbstractTableModel):
     def __init__(self, die, prefix, lowlevel, hex, regnames):
         QAbstractTableModel.__init__(self)
@@ -125,8 +41,10 @@ class DIETableModel(QAbstractTableModel):
         self.keys = list(die.attributes.keys())
         self.headers = _ll_headers if self.lowlevel else _noll_headers
         self.meta_count = _meta_count if lowlevel else 0
-        self.arch = die.dwarfinfo.config.machine_arch
-        self.regnamelist = _REG_NAME_MAP.get(self.arch)
+        self.expr_formatter = ExprFormatter(regnames, prefix, die.dwarfinfo.config.machine_arch, die.cu['version'], hex)
+
+    from .ranges import show_ranges
+    from .locs import parse_location, show_location 
 
     def headerData(self, section, ori, role):
         if ori == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
@@ -138,21 +56,23 @@ class DIETableModel(QAbstractTableModel):
     def columnCount(self, parent):
         return len(self.headers)
 
-    def parse_location(self, attr):
-        di = self.die.dwarfinfo
-        if di._locparser is None:
-            di._locparser = LocationParser(di.location_lists())
-        return di._locparser.parse_from_attribute(attr, self.die.cu['version'], die = self.die)
-
     def data(self, index, role):
         row = index.row()
         return self.attr_data(index, role) if row >= self.meta_count else self.meta_data(index, role)
+    
+    # End of Qt callbacks
+
+    # Flavors of data() - attribute and metadata
 
     def attr_data(self, index, role):
         irow = index.row()
         meta_count = self.meta_count
         self_keys_len = len(self.keys)
         self_die_keys_len = len(self.die.attributes.keys())
+
+        if irow >= meta_count + self_keys_len: # Shorting out 1530
+            return None
+                
         row = index.row() - self.meta_count
         key = self.keys[row]
         attr = self.attributes[key]
@@ -202,47 +122,6 @@ class DIETableModel(QAbstractTableModel):
         elif role == Qt.ItemDataRole.BackgroundRole:
             return _ltgrey_brush
 
-    # End of Qt callbacks
-
-    def format_arg(self, s):
-        if isinstance(s, str):
-            return s
-        elif isinstance(s, int):
-            return hex(s) if self.hex else str(s)
-        elif isinstance(s, list): # Could be a blob (list of ints), could be a subexpression
-            if len(s) > 0 and isinstance(s[0], DWARFExprOp): # Subexpression
-                return '{' + '; '.join(self.format_op(*op) for op in s) + '}'
-            else:
-                return bytes(s).hex() # Python 3.5+
-
-    def decode_breg(self, regno, offset):
-        if offset < 0:
-            return '[%s-0x%x]' % (self.regnamelist[regno], -offset)
-        else:
-            return '[%s+0x%x]' % (self.regnamelist[regno], offset)
-
-    def format_op(self, op, op_name, args, offset):
-        if not self.regnames and self.regnamelist: # Friendly register names
-            if 0x50 <= op <= 0x6f and op - 0x50 < len(self.regnamelist): # reg0...reg31
-                op_name = self.regnamelist[op-0x50]
-            elif 0x70 <= op <= 0x8f and op - 0x70 < len(self.regnamelist) and len(args) > 0: # breg0...breg31(offset)
-                op_name = self.decode_breg(op - 0x70, args[0])
-                args = False
-            elif (op == 0x90 or (self.die.cu['version'] == 1 and op == 0x1)) and len(args) > 0 and args[0] >= 0 and args[0] < len(self.regnamelist): # regx(regno)
-                op_name = self.regnamelist[args[0]]
-                args = False
-            elif op == 0x92 and len(args) > 1 and args[0] >= 0 and args[0] < len(self.regnamelist): # bregx(regno, offset)
-                op_name = self.decode_breg(args[0], args[1])
-                args = False
-
-        if op_name.startswith('DW_OP_') and not self.prefix:
-            op_name = op_name[6:]
-
-        if args:
-            return op_name + ' ' + ', '.join(self.format_arg(s) for s in args)
-        else:
-            return op_name
-
     # Expr is an expression blob
     # Returns a list of strings for ops
     # Format: op arg, arg...
@@ -253,15 +132,15 @@ class DIETableModel(QAbstractTableModel):
         # Challenge: for nested expressions, args is a list with a list of commands
         # For those, the format is: op {op arg, arg; op arg, arg}
         # Can't just check for iterable, str is iterable too
-        return [self.format_op(*op) for op in self.die.cu._exprparser.parse_expr(expr)]
+        return [self.expr_formatter.format_op(*op) for op in self.die.cu._exprparser.parse_expr(expr)]
 
-    # Big DIE attribute value interpreter
+    # Big DIE attribute value interpreter for the top right table
     def format_value(self, attr):
         try:
             die = self.die
             cu = self.die.cu
             header = self.die.cu.header
-            dwarf_version = self.die.cu.header.version                               
+            dwarf_version = self.die.cu.header.version
 
             key = attr.name
             val = attr.value
@@ -276,7 +155,9 @@ class DIETableModel(QAbstractTableModel):
                 return str(bool(val))
             elif LocationParser.attribute_has_location(attr, self.die.cu['version']):
                 ll = self.parse_location(attr)
-                if isinstance(ll, LocationExpr):
+                if ll is None:
+                    return "(parse error - please report at github.com/sevaa/dwex)"
+                elif isinstance(ll, LocationExpr):
                     return '; '.join(self.dump_expr(ll.loc_expr))
                 else:
                     return "Loc list: 0x%x" % attr.value
@@ -288,6 +169,8 @@ class DIETableModel(QAbstractTableModel):
                 return "%d %s" % (val, _DESCR_DW_ACCESS[val]) if val in _DESCR_DW_ACCESS else val
             elif key == 'DW_AT_inline':
                 return "%d %s" % (val, _DESCR_DW_INL[val]) if val in _DESCR_DW_INL else val
+            elif key == 'DW_AT_calling_convention':
+                return "%d %s" % (val, _DESCR_DW_CC[val]) if val in _DESCR_DW_CC else val
             elif key in ('DW_AT_decl_file', 'DW_AT_call_file'):
                 cu = self.die.cu
                 if cu._lineprogram is None:
@@ -329,11 +212,25 @@ class DIETableModel(QAbstractTableModel):
                     return str(val)
             else:
                 return hex(val) if self.hex and isinstance(val, int) else str(val)
+        # except BaseException as exc:
         except ELFParseError as exc:
-            from .__main__ import version
-            from .crash import report_crash
-            tb = exc.__traceback__
-            report_crash(exc, tb, version, ctxt = {'attr': attr, 'die':die, 'cu_header':header, 'dwarf_version':dwarf_version})
+            import sys
+            if not sys.gettrace():
+                from .__main__ import version
+                from .crash import report_crash
+                from inspect import currentframe
+                tb = exc.__traceback__
+                di = die.cu.dwarfinfo
+                file_addr_size = di.config.default_address_size
+                loc_section = di.debug_loclists_sec if dwarf_version >= 5 else di.debug_loc_sec
+                loc_sec_len = len(loc_section.stream.getbuffer()) if loc_section else None
+                ctxt = {'attr': attr,
+                        'die': die,
+                        'cu_header': header,
+                        'file_addr_size': file_addr_size,
+                        'dwarf_version': dwarf_version,
+                        'sec_size': loc_sec_len}
+                report_crash(exc, tb, version, currentframe(), ctxt)
             return "(parse error - please report at github.com/sevaa/dwex)"
 
     def format_form(self, form):
@@ -355,9 +252,9 @@ class DIETableModel(QAbstractTableModel):
         self.die = die
         self.attributes = die.attributes
         self.keys = list(die.attributes.keys())
-        if die.dwarfinfo.config.machine_arch != self.arch:
-            self.arch = die.dwarfinfo.config.machine_arch
-            self.regnamelist = _REG_NAME_MAP.get(self.arch)        
+        # Should not be ever possible, but this code is here
+        self.expr_formatter.set_arch(die.dwarfinfo.config.machine_arch)
+        self.expr_formatter.dwarf_version = die.cu['version']
         rows_now = self.meta_count + len(self.keys)
         if rows_was < rows_now:
             self.rowsInserted.emit(QModelIndex(), rows_was, rows_now-1)
@@ -368,6 +265,7 @@ class DIETableModel(QAbstractTableModel):
     def set_prefix(self, prefix):
         if prefix != self.prefix:
             self.prefix = prefix
+            self.expr_formatter.prefix = prefix
             self.refresh_values()
 
     # Index is the current selected index
@@ -408,6 +306,7 @@ class DIETableModel(QAbstractTableModel):
     def set_regnames(self, regnames):
         if regnames != self.regnames:
             self.regnames = regnames
+            self.expr_formatter.regnames = regnames
             self.refresh_values()
 
     # Returns a table model for the attribute details table
@@ -421,64 +320,9 @@ class DIETableModel(QAbstractTableModel):
             attr = self.attributes[key]
             form = attr.form
             if key == "DW_AT_ranges":
-                di = self.die.dwarfinfo
-                if not di._ranges:
-                    di._ranges = di.range_lists()
-                if not di._ranges: # Absent in the DWARF file
-                    return None
-                ranges = di._ranges.get_range_list_at_offset(attr.value, cu = self.die.cu)
-                top_die = self.die.cu.get_top_DIE() 
-                warn = None
-                lines = []                
-                if len(ranges):
-                    cu_base = 0
-                    # Do we need the base address? We might not.
-                    has_relative_entries = next((r for r in ranges if isinstance(r, RangeEntry) and not r.is_absolute), False)
-                    if has_relative_entries and not isinstance(ranges[0], RangeBaseAddressEntry):
-                        try:
-                            cu_base = get_cu_base(self.die)
-                        except NoBaseError:
-                            warn = "Base address not found, assuming 0"
-
-                    # TODO: low level view?
-                    for r in ranges:
-                        if isinstance(r, RangeEntry):
-                            base = 0 if r.is_absolute else cu_base
-                            lines.append((hex(base + r.begin_offset), hex(base + r.end_offset)))
-                        else:
-                            cu_base = r.base_address
-                else:
-                    warn = "Empty range list"
-                return GenericTableModel(("Start offset", "End offset"), lines, warn)
+                return self.show_ranges(attr)
             elif LocationParser.attribute_has_location(attr, self.die.cu['version']):
-                # Expression is a list of ints
-                ll = self.parse_location(attr)
-                if isinstance(ll, LocationExpr):
-                    return GenericTableModel(("Command",), ((cmd,) for cmd in self.dump_expr(ll.loc_expr)))
-                else:
-                    cu_base = get_cu_base(self.die)
-                    values = list()
-                    if self.lowlevel:
-                        headers = ("Start offset", "End offset", "Expr bytes", "Expression")
-                        for l in ll:
-                            if 'base_address' in l._fields:
-                                cu_base = l.base_address
-                                values.append(("(base)", hex(cu_base), '', ''))
-                            else:
-                                values.append((hex(cu_base + l.begin_offset),
-                                    hex(cu_base + l.end_offset),
-                                    ' '.join("%02x" % b for b in l.loc_expr),
-                                    '; '.join(self.dump_expr(l.loc_expr))))
-                    else:
-                        headers = ("Start offset", "End offset", "Expression")
-                        for l in ll:
-                            if 'base_address' in l._fields:
-                                cu_base = l.base_address
-                            else:
-                                values.append((hex(cu_base + l.begin_offset),
-                                    hex(cu_base + l.end_offset),
-                                    '; '.join(self.dump_expr(l.loc_expr))))                        
-                    return GenericTableModel(headers, values)
+                return self.show_location(attr)
             elif key == 'DW_AT_stmt_list':
                 if self.die.cu._lineprogram is None:
                     self.die.cu._lineprogram = self.die.dwarfinfo.line_program_for_CU(self.die.cu)
@@ -538,69 +382,9 @@ class DIETableModel(QAbstractTableModel):
         except IndexError as exc:
             from .__main__ import version
             from .crash import report_crash
+            from inspect import currentframe
             tb = exc.__traceback__
-            report_crash(exc, tb, version)
+            report_crash(exc, tb, version, currentframe())
             return None
 
-class GenericTableModel(QAbstractTableModel):
-    def __init__(self, headers, values, warning = None):
-        QAbstractTableModel.__init__(self)
-        self.headers = headers
-        self.values = tuple(values)
-        self.warning = warning
 
-    def headerData(self, section, ori, role):
-        if ori == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
-            return self.headers[section]
-
-    def rowCount(self, parent):
-        return len(self.values)
-
-    def columnCount(self, parent):
-        return len(self.headers)
-
-    def data(self, index, role):
-        if role == Qt.ItemDataRole.DisplayRole:
-            return self.values[index.row()][index.column()]
-
-class FixedWidthTableModel(GenericTableModel):
-    def __init__(self, headers, values):
-        GenericTableModel.__init__(self, headers, values)
-
-    def data(self, index, role):
-        if role == Qt.ItemDataRole.FontRole:
-            global _fixed_font
-            if not _fixed_font:
-                _fixed_font = QFont("Monospace")
-                _fixed_font.setStyleHint(QFont.StyleHint.TypeWriter)
-            return _fixed_font
-        else:
-            return GenericTableModel.data(self, index, role)
-
-
-# Find helper:
-# Returns true if the specified IP is in [low_pc, high_pc)
-# Or in ranges
-def ip_in_range(die, ip):
-    if 'DW_AT_ranges' in die.attributes:
-        di = die.dwarfinfo
-        if not di._ranges:
-            di._ranges = di.range_lists()
-        if not di._ranges: # Absent in the DWARF file
-            return False
-        # TODO: handle base addresses. Never seen those so far...
-        cu_base = get_cu_base(die)
-        rl = di._ranges.get_range_list_at_offset(die.attributes['DW_AT_ranges'].value, cu = die.cu)
-        for r in rl:
-            if isinstance(r, RangeBaseAddressEntry):
-                cu_base = r.base_address
-            elif r.begin_offset <= ip - cu_base < r.end_offset:
-                return True
-    if 'DW_AT_low_pc' in die.attributes and 'DW_AT_high_pc' in die.attributes:
-        l = die.attributes['DW_AT_low_pc'].value
-        h = die.attributes['DW_AT_high_pc'].value
-        if die.attributes['DW_AT_high_pc'].form != 'DW_FORM_addr':
-            h += l
-        if l <= ip < h:
-            return True
-    return False
