@@ -2,6 +2,7 @@ from bisect import bisect_left, bisect_right
 from PyQt6.QtCore import Qt, QAbstractItemModel, QModelIndex
 from PyQt6.QtGui import QFont, QFontInfo, QBrush
 from PyQt6.QtWidgets import QApplication, QMessageBox
+from .dwarfutil import has_code_location, safe_DIE_name
 
 # Supports both / and \ - current system separator might not match the system the file came from
 # so os.path.basename won't do
@@ -28,10 +29,7 @@ def cu_sort_key(cu):
     return top_die_file_name(cu.get_top_DIE()).lower()
 
 def die_sort_key(die):
-    if 'DW_AT_name' in die.attributes:
-        name = die.attributes['DW_AT_name'].value.decode('utf-8', errors='ignore').lower()
-    else:
-        name = ''
+    name = safe_DIE_name(die)
     tag = '%X' % die.tag if isinstance(die.tag, int) else die.tag
     return (tag, name, die.offset)
 
@@ -49,9 +47,9 @@ def load_children(parent_die, sort):
     # Load and cache child DIEs in the parent DIE, if necessary
     # Assumes the check if the DIE has children has been already performed
     if not hasattr(parent_die, "_children") or parent_die._children is None:
-        # TODO: wait cursor here. It may cause disk I/O
+        # TODO: wait cursor here.
         try:
-            parent_die._children = [decorate_die(die, i) for (i, die) in enumerate(parent_die.iter_children())]    
+            parent_die._children = [decorate_die(die, i) for (i, die) in enumerate(parent_die.iter_children())]
             if sort:
                 parent_die._children.sort(key = die_sort_key)
                 for (i, die) in enumerate(parent_die._children):
@@ -133,18 +131,40 @@ class DWARFTreeModel(QAbstractItemModel):
                 return s
         elif role == Qt.ItemDataRole.ToolTipRole:
             if die.tag == 'DW_TAG_compile_unit' or die.tag == 'DW_TAG_partial_unit':
-                return die.attributes['DW_AT_name'].value.decode('utf-8', errors='ignore') if 'DW_AT_name' in die.attributes else None
-        elif role == Qt.ItemDataRole.ForegroundRole and self.highlight_condition and self.highlight_condition(die):
+                return safe_DIE_name(die, None)
+        elif role == Qt.ItemDataRole.ForegroundRole and self.is_highlighted(die):
             return self.blue_brush
-        elif role == Qt.ItemDataRole.FontRole and self.highlight_condition and self.highlight_condition(die):
+        elif role == Qt.ItemDataRole.FontRole and self.is_highlighted(die):
             return self.bold_font
 
     # The rest is not Qt callbacks
 
-    def highlight(self, condition):
-        self.highlight_condition = condition
+    def is_highlighted(self, die):
+        if not self.highlight_condition:
+            return False
+        return next((True for v in self.highlight_condition.values() if v(die)), False)
+
+    def add_highlight(self, key, condition):
+        if not self.highlight_condition:
+            self.highlight_condition = {key:condition}
+        else:
+            self.highlight_condition[key] = condition
         self.dataChanged.emit(self.createIndex(0, 0, self.top_dies[0]), self.createIndex(len(self.top_dies)-1, 0, self.top_dies[-1]), (Qt.ItemDataRole.ForegroundRole, Qt.ItemDataRole.FontRole))
 
+    def remove_highlight(self, key):
+        del self.highlight_condition[key]
+        self.dataChanged.emit(self.createIndex(0, 0, self.top_dies[0]), self.createIndex(len(self.top_dies)-1, 0, self.top_dies[-1]), (Qt.ItemDataRole.ForegroundRole, Qt.ItemDataRole.FontRole))
+
+    def clear_highlight(self):
+        self.highlight_condition = None
+        self.dataChanged.emit(self.createIndex(0, 0, self.top_dies[0]), self.createIndex(len(self.top_dies)-1, 0, self.top_dies[-1]), (Qt.ItemDataRole.ForegroundRole, Qt.ItemDataRole.FontRole))
+
+    def has_highlight(self, key):
+        return bool(self.highlight_condition and key in self.highlight_condition)
+    
+    def has_any_highlights(self):
+        return bool(self.highlight_condition and len(self.highlight_condition))
+
     def set_prefix(self, prefix):
         if prefix != self.prefix:
             self.prefix = prefix
@@ -205,7 +225,7 @@ class DWARFTreeModel(QAbstractItemModel):
 
         # Abusing the structure of the per-CU DIE cache of pyelftools, it's the same in DWWARFv1
         i = bisect_left(target_cu._diemap, target_offset)
-        if target_cu._diemap[i] != target_offset:
+        if i >= len(target_cu._diemap) or target_cu._diemap[i] != target_offset:
             return None
         target_die = target_cu._dielist[i]
         if target_die.is_null():
@@ -214,7 +234,7 @@ class DWARFTreeModel(QAbstractItemModel):
 
     # Takes a die that might not have an _i
     # Restores the _i
-    # Assumes all parent DIEs of the current one are already parsed
+    # Assumes some parent DIEs of the current one are already parsed
     # and cached in the CU, so get_parent will always return a valid parent DIE
     def index_for_die(self, die):
         if hasattr(die, '_i'): # DIE already iterated over
@@ -225,6 +245,9 @@ class DWARFTreeModel(QAbstractItemModel):
                 parent_die = die.get_parent()
                 load_children(parent_die, self.sortdies) # This will populate the _i in all children of parent_die, including die
                 if not index: # After the first iteration, the one in the direct parent of target_die, target_die will have _i
+                    if die.is_null():
+                        die = parent_die._children[-1] # Null is a terminator in a sequence - move to the sibling
+                        # TODO: move to the closest in terms of offset, which would require going down the nearest sibling's tree
                     index = self.createIndex(die._i, 0, die)
                 die = parent_die
             return index
@@ -253,7 +276,7 @@ class DWARFTreeModel(QAbstractItemModel):
                     # Quit condition with search from position - quit once we go past the starting position after the wrap
                     if have_start_pos and cu_offset >= start_cu_offset and die.offset > start_die_offset and wrapped:
                         break
-                    if (not have_start_pos or cu_offset != start_cu_offset or (not wrapped and die.offset > start_die_offset)) and cond(die):
+                    if not die.is_null() and (not have_start_pos or cu_offset != start_cu_offset or (not wrapped and die.offset > start_die_offset)) and cond(die):
                         return self.index_for_die(die)
 
             # We're at the end of the CU. What next?
@@ -266,6 +289,55 @@ class DWARFTreeModel(QAbstractItemModel):
                 break
 
         return False
+
+    # Search back - same idea
+    def find_back(self, start_pos, cond, cu_cond = False):
+        have_start_pos = start_pos.isValid()
+        if have_start_pos: # Searching from a specific position, with wrap-around
+            start_die = start_pos.internalPointer()
+            start_die_offset = start_die.offset # In the current die, before the next one
+            start_cu = start_die.cu
+            start_cu_offset = start_cu.cu_offset
+            cu = start_cu
+            wrapped = False
+        else:
+            cu = self.top_dies[-1].cu
+
+        while True:
+            cu_offset = cu.cu_offset
+            # Parse all DIEs in the current CU
+            if cu_cond(cu) if cu_cond else True:
+                for die in cu.iter_DIEs(): # Fill the DIE cache
+                    pass
+
+                # Abusing the internal cache of pyelftools - fragile!
+
+                if have_start_pos and not wrapped and cu_offset == start_cu_offset:
+                    i = bisect_left(cu._diemap, start_die_offset-1)-1
+                else:
+                    i = len(cu._diemap) - 1
+
+                while i >= 0:
+                    die = cu._dielist[i]
+                    if not die.is_null():
+                        # Quit condition with search from position - quit once we go past the starting position after the wrap
+                        if have_start_pos and die.offset == start_die_offset and wrapped:
+                            return False
+                        if cond(die):
+                            return self.index_for_die(die)
+                    i -= 1
+
+            # We're at the end of the CU. What next?
+            if cu._i > 0: # More CUs to scan
+                cu = self.top_dies[cu._i - 1].cu
+            elif have_start_pos and not wrapped: # Scanned the last CU, wrap around
+                cu = self.top_dies[-1].cu
+                wrapped = True
+            else:
+                break
+
+        return False            
+
     
     def find_offset(self, offset):
         cu = next((td.cu
@@ -286,7 +358,3 @@ class DWARFTreeModel(QAbstractItemModel):
         i = bisect_right(cu._diemap, offset)
         return self.index_for_die(cu._dielist[i - 1])
 
-# Highlighter function(s)
-def has_code_location(die):
-    return 'DW_AT_low_pc' in die.attributes or 'DW_AT_ranges' in die.attributes
-
