@@ -1,4 +1,5 @@
 from testall import test_file, test_tree, test_tree_for, test_file_for
+from dwex.patch import monkeypatch
 import os
 
 def save_info():
@@ -16,12 +17,19 @@ def dump_compiler(di):
     die = next(di.iter_CUs()).get_top_DIE()
     print(die.attributes['DW_AT_producer'].value.decode('UTF-8'))
 
-#test_file("samples\\sles_507.88")
-test_tree("R:\\pyelftools2\\test\\testfiles_for_readelf")
+monkeypatch()
+
+def test_CU_iter(di):
+    for cu in di.iter_CUs():
+        for die in cu.iter_DIEs():
+            pass
+
 #test_tree_for("\\\\sandbox\\seva\\webdata\\clarify\\derived", dump_compiler)    
 #test_tree_for("\\\\mac\\seva\\quincy", dump_compiler)
 
-#test_tree("samples")
+#test_tree("E:\\Seva\\Projects\\pyelftools\\test\\testfiles_for_readelf")
+#test_tree("E:\\Seva\\Projects\\pyelftools\\test\\testfiles_for_dwarfdump")
+test_tree("samples\\other")
 
 def f(di):
     CUs = [cu for cu in di.iter_CUs()]
